import datetime
import os
import random
import string
import sys
import unittest2

from mock import patch, Mock

import stripe

NOW = datetime.datetime.now()

DUMMY_CHARGE = {
    'amount': 100,
    'currency': 'usd',
    'source': 'tok_visa'
}

DUMMY_DISPUTE = {
    'status': 'needs_response',
    'currency': 'usd',
    'metadata': {}
}

DUMMY_PLAN = {
    'amount': 2000,
    'interval': 'month',
    'name': 'Amazing Gold Plan',
    'currency': 'usd',
    'id': ('stripe-test-gold-' +
           ''.join(random.choice(string.ascii_lowercase) for x in range(10)))
}

DUMMY_COUPON = {
    'percent_off': 25,
    'duration': 'repeating',
    'duration_in_months': 5,
    'metadata': {}
}

DUMMY_RECIPIENT = {
    'name': 'John Doe',
    'type': 'individual'
}

DUMMY_TRANSFER = {
    'amount': 400,
    'currency': 'usd',
    'recipient': 'self'
}

DUMMY_APPLE_PAY_DOMAIN = {
    'domain_name': 'test.com',
}

DUMMY_INVOICE_ITEM = {
    'amount': 456,
    'currency': 'usd',
}

SAMPLE_INVOICE = stripe.util.json.loads("""
{
  "amount_due": 1305,
  "attempt_count": 0,
  "attempted": true,
  "charge": "ch_wajkQ5aDTzFs5v",
  "closed": true,
  "customer": "cus_osllUe2f1BzrRT",
  "date": 1338238728,
  "discount": null,
  "ending_balance": 0,
  "id": "in_t9mHb2hpK7mml1",
  "livemode": false,
  "next_payment_attempt": null,
  "object": "invoice",
  "paid": true,
  "period_end": 1338238728,
  "period_start": 1338238716,
  "starting_balance": -8695,
  "subtotal": 10000,
  "total": 10000,
  "lines": {
    "invoiceitems": [],
    "prorations": [],
    "subscriptions": [
      {
        "plan": {
          "interval": "month",
          "object": "plan",
          "identifier": "expensive",
          "currency": "usd",
          "livemode": false,
          "amount": 10000,
          "name": "Expensive Plan",
          "trial_period_days": null,
          "id": "expensive"
        },
        "period": {
          "end": 1340917128,
          "start": 1338238728
        },
        "amount": 10000
      }
    ]
  }
}
""")

DUMMY_WEBHOOK_PAYLOAD = """{
  "id": "evt_test_webhook",
  "object": "event"
}"""

DUMMY_WEBHOOK_SECRET = 'whsec_test_secret'


class StripeTestCase(unittest2.TestCase):
    RESTORE_ATTRIBUTES = ('api_version', 'api_key', 'client_id')

    def setUp(self):
        super(StripeTestCase, self).setUp()

        self._stripe_original_attributes = {}

        for attr in self.RESTORE_ATTRIBUTES:
            self._stripe_original_attributes[attr] = getattr(stripe, attr)

        api_base = os.environ.get('STRIPE_API_BASE')
        if api_base:
            stripe.api_base = api_base
        stripe.api_key = os.environ.get(
            'STRIPE_API_KEY', 'tGN0bIwXnHdwOa85VABjPdSn8nWY7G7I')
        stripe.api_version = os.environ.get(
            'STRIPE_API_VERSION', '2017-04-06')

    def tearDown(self):
        super(StripeTestCase, self).tearDown()

        for attr in self.RESTORE_ATTRIBUTES:
            setattr(stripe, attr, self._stripe_original_attributes[attr])


class StripeUnitTestCase(StripeTestCase):
    REQUEST_LIBRARIES = ['urlfetch', 'requests', 'pycurl']

    if sys.version_info >= (3, 0):
        REQUEST_LIBRARIES.append('urllib.request')
    else:
        REQUEST_LIBRARIES.append('urllib2')

    def setUp(self):
        super(StripeUnitTestCase, self).setUp()

        self.request_patchers = {}
        self.request_mocks = {}
        for lib in self.REQUEST_LIBRARIES:
            patcher = patch("stripe.http_client.%s" % (lib,))

            self.request_mocks[lib] = patcher.start()
            self.request_patchers[lib] = patcher

    def tearDown(self):
        super(StripeUnitTestCase, self).tearDown()

        for patcher in self.request_patchers.itervalues():
            patcher.stop()


class StripeApiTestCase(StripeTestCase):

    def setUp(self):
        super(StripeApiTestCase, self).setUp()

        self.requestor_patcher = patch('stripe.api_requestor.APIRequestor')
        self.requestor_class_mock = self.requestor_patcher.start()
        self.requestor_mock = self.requestor_class_mock.return_value

    def tearDown(self):
        super(StripeApiTestCase, self).tearDown()

        self.requestor_patcher.stop()

    def mock_response(self, res):
        self.requestor_mock.request = Mock(return_value=(res, 'reskey'))


class StripeResourceTest(StripeApiTestCase):

    def setUp(self):
        super(StripeResourceTest, self).setUp()
        self.mock_response({})


class MyResource(stripe.resource.APIResource):
    pass


class MySingleton(stripe.resource.SingletonAPIResource):
    pass


class MyListable(stripe.resource.ListableAPIResource):
    pass


class MyCreatable(stripe.resource.CreateableAPIResource):
    pass


class MyUpdateable(stripe.resource.UpdateableAPIResource):
    pass


class MyDeletable(stripe.resource.DeletableAPIResource):
    pass


class MyComposite(stripe.resource.ListableAPIResource,
                  stripe.resource.CreateableAPIResource,
                  stripe.resource.UpdateableAPIResource,
                  stripe.resource.DeletableAPIResource):
    pass



# Stripe Python bindings
# API docs at http://stripe.com/docs/api
# Authors:
# Patrick Collison <patrick@stripe.com>
# Greg Brockman <gdb@stripe.com>
# Andrew Metcalf <andrew@stripe.com>

# Configuration variables

api_key = None
client_id = None
api_base = 'https://api.stripe.com'
connect_api_base = 'https://connect.stripe.com'
upload_api_base = 'https://uploads.stripe.com'
api_version = None
verify_ssl_certs = True
proxy = None
default_http_client = None
app_info = None

# Set to either 'debug' or 'info', controls console logging
log = None

# Resource
from stripe.resource import (  # noqa
    Account,
    AlipayAccount,
    ApplePayDomain,
    ApplicationFee,
    Balance,
    BalanceTransaction,
    BankAccount,
    BitcoinReceiver,
    BitcoinTransaction,
    Card,
    Charge,
    CountrySpec,
    Coupon,
    Customer,
    Dispute,
    EphemeralKey,
    Event,
    FileUpload,
    Invoice,
    InvoiceItem,
    LoginLink,
    Order,
    OrderReturn,
    Payout,
    Plan,
    Product,
    Recipient,
    RecipientTransfer,
    Refund,
    SKU,
    Source,
    Subscription,
    SubscriptionItem,
    ThreeDSecure,
    Token,
    Transfer)

# OAuth
from stripe.oauth import OAuth  # noqa

# Webhooks
from stripe.webhook import Webhook, WebhookSignature  # noqa

# Error imports.  Note that we may want to move these out of the root
# namespace in the future and you should prefer to access them via
# the fully qualified `stripe.error` module.

from stripe.error import (  # noqa
    APIConnectionError,
    APIError,
    AuthenticationError,
    PermissionError,
    RateLimitError,
    CardError,
    InvalidRequestError,
    SignatureVerificationError,
    StripeError)

# OAuth error classes are not imported into the root namespace and must be
# accessed via stripe.oauth_error.<Exception>
from stripe import oauth_error  # noqa

# DEPRECATED: These imports will be moved out of the root stripe namespace
# in version 2.0

from stripe.version import VERSION  # noqa
from stripe.api_requestor import APIRequestor  # noqa
from stripe.resource import (  # noqa
    APIResource,
    CreateableAPIResource,
    DeletableAPIResource,
    ListObject,
    ListableAPIResource,
    SingletonAPIResource,
    StripeObject,
    StripeObjectEncoder,
    UpdateableAPIResource,
    convert_to_stripe_object)
from stripe.util import json, logger  # noqa


# Sets some basic information about the running application that's sent along
# with API requests. Useful for plugin authors to identify their plugin when
# communicating with Stripe.
#
# Takes a name and optional version and plugin URL.
def set_app_info(name, version=None, url=None):
    global app_info
    app_info = {
        'name': name,
        'version': version,
        'url': url,
    }



import stripe
from stripe.test.helper import (StripeResourceTest)


class RecipientTest(StripeResourceTest):

    def test_list_recipients(self):
        stripe.Recipient.list()
        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/recipients',
            {}
        )

    def test_recipient_transfers(self):
        recipient = stripe.Recipient(id='rp_transfer')
        recipient.transfers()

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/transfers',
            {'recipient': 'rp_transfer'},
        )

    def test_recipient_add_card(self):
        recipient = stripe.Recipient.construct_from({
            'id': 'rp_add_card',
            'sources': {
                'object': 'list',
                'url': '/v1/recipients/rp_add_card/sources',
            },
        }, 'api_key')
        recipient.sources.create(card='tok_visa_debit')

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/recipients/rp_add_card/sources',
            {
                'card': 'tok_visa_debit',
            },
            None
        )

    def test_recipient_update_card(self):
        card = stripe.Card.construct_from({
            'recipient': 'rp_update_card',
            'id': 'ca_update_card',
        }, 'api_key')
        card.name = 'The Best'
        card.save()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/recipients/rp_update_card/cards/ca_update_card',
            {
                'name': 'The Best',
            },
            None
        )

    def test_recipient_delete_card(self):
        card = stripe.Card.construct_from({
            'recipient': 'rp_delete_card',
            'id': 'ca_delete_card',
        }, 'api_key')
        card.delete()

        self.requestor_mock.request.assert_called_with(
            'delete',
            '/v1/recipients/rp_delete_card/cards/ca_delete_card',
            {},
            None
        )



import stripe
from stripe.test.helper import (
    StripeResourceTest, DUMMY_PLAN
)


class SubscriptionItemTest(StripeResourceTest):

    def test_list_subscriptions(self):
        stripe.SubscriptionItem.all(subscription='test_sub', limit=3)
        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/subscription_items',
            {
                'subscription': 'test_sub',
                'limit': 3,
            },
        )

    def test_retrieve_subscription(self):
        stripe.SubscriptionItem.retrieve('test_sub_item')

        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/subscription_items/test_sub_item',
            {},
            None
        )

    def test_create_subscription(self):
        stripe.SubscriptionItem.create(subscription='test_sub',
                                       plan=DUMMY_PLAN['id'])

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/subscription_items',
            {
                'subscription': 'test_sub',
                'plan': DUMMY_PLAN['id']
            },
            None
        )

    def test_update_subscription(self):
        item = stripe.SubscriptionItem.construct_from({
            'id': 'test_sub_item',
            'plan': 'test_plan',
        }, 'api_key')

        item.plan = DUMMY_PLAN['id']
        item.save()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/subscription_items/test_sub_item',
            {
                'plan': DUMMY_PLAN['id'],
            },
            None
        )

    def test_modify_subscription(self):
        stripe.SubscriptionItem.modify('test_sub_item',
                                       plan=DUMMY_PLAN['id'])

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/subscription_items/test_sub_item',
            {
                'plan': DUMMY_PLAN['id'],
            },
            None
        )

    def test_delete_subscription(self):
        item = stripe.SubscriptionItem.construct_from({
            'id': 'test_sub_item',
            'plan': 'test_plan',
        }, 'api_key')

        item.delete()

        self.requestor_mock.request.assert_called_with(
            'delete',
            '/v1/subscription_items/test_sub_item',
            {},
            None
        )



import urlparse

import stripe
from stripe.test.helper import StripeApiTestCase


class OAuthTests(StripeApiTestCase):
    def setUp(self):
        super(OAuthTests, self).setUp()
        self.mock_response({})

        stripe.client_id = 'ca_test'

    def tearDown(self):
        super(OAuthTests, self).tearDown()

        stripe.client_id = None

    def test_authorize_url(self):
        url = stripe.OAuth.authorize_url(
            scope='read_write',
            state='csrf_token',
            stripe_user={
                'email': 'test@example.com',
                'url': 'https://example.com/profile/test',
                'country': 'US',
            })

        o = urlparse.urlparse(url)
        params = urlparse.parse_qs(o.query)

        self.assertEqual('https', o.scheme)
        self.assertEqual('connect.stripe.com', o.netloc)
        self.assertEqual('/oauth/authorize', o.path)

        self.assertEqual(['ca_test'], params['client_id'])
        self.assertEqual(['read_write'], params['scope'])
        self.assertEqual(['test@example.com'], params['stripe_user[email]'])
        self.assertEqual(
            ['https://example.com/profile/test'],
            params['stripe_user[url]']
        )
        self.assertEqual(['US'], params['stripe_user[country]'])

    def test_token(self):
        stripe.OAuth.token(
            grant_type='authorization_code',
            code='this_is_an_authorization_code',
        )

        self.requestor_mock.request.assert_called_with(
            'post',
            '/oauth/token',
            {
                'grant_type': 'authorization_code',
                'code': 'this_is_an_authorization_code',
            },
            None
        )

    def test_deauthorize(self):
        stripe.OAuth.deauthorize(stripe_user_id='acct_deauth')

        self.requestor_mock.request.assert_called_with(
            'post',
            '/oauth/deauthorize',
            {
                'client_id': 'ca_test',
                'stripe_user_id': 'acct_deauth',
            },
            None
        )

    def test_deauthorize_account_instance(self):
        acct = stripe.Account.construct_from({
            'id': 'acct_deauth',
        }, 'api_key')
        acct.deauthorize()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/oauth/deauthorize',
            {
                'client_id': 'ca_test',
                'stripe_user_id': 'acct_deauth',
            },
            None
        )



import stripe
from stripe.test.helper import (
    StripeApiTestCase, MyResource, MyUpdateable, MySingleton
)


class APIResourceTests(StripeApiTestCase):

    def test_retrieve_and_refresh(self):
        self.mock_response({
            'id': 'foo2',
            'bobble': 'scrobble',
        })

        res = MyResource.retrieve('foo*', myparam=5)

        url = '/v1/myresources/foo%2A'
        self.requestor_mock.request.assert_called_with(
            'get', url, {'myparam': 5}, None
        )

        self.assertEqual('scrobble', res.bobble)
        self.assertEqual('foo2', res.id)
        self.assertEqual('reskey', res.api_key)

        self.mock_response({
            'frobble': 5,
        })

        res = res.refresh()

        url = '/v1/myresources/foo2'
        self.requestor_mock.request.assert_called_with(
            'get', url, {'myparam': 5}, None
        )

        self.assertEqual(5, res.frobble)
        self.assertRaises(KeyError, res.__getitem__, 'bobble')

    def test_convert_to_stripe_object(self):
        sample = {
            'foo': 'bar',
            'adict': {
                'object': 'charge',
                'id': 42,
                'amount': 7,
            },
            'alist': [
                {
                    'object': 'customer',
                    'name': 'chilango'
                }
            ]
        }

        converted = stripe.resource.convert_to_stripe_object(
            sample, 'akey', None, None)

        # Types
        self.assertTrue(isinstance(converted, stripe.resource.StripeObject))
        self.assertTrue(isinstance(converted.adict, stripe.Charge))
        self.assertEqual(1, len(converted.alist))
        self.assertTrue(isinstance(converted.alist[0], stripe.Customer))

        # Values
        self.assertEqual('bar', converted.foo)
        self.assertEqual(42, converted.adict.id)
        self.assertEqual('chilango', converted.alist[0].name)

        # Stripping
        # TODO: We should probably be stripping out this property
        # self.assertRaises(AttributeError, getattr, converted.adict, 'object')

    def test_convert_array_to_dict(self):
        out = stripe.resource.convert_array_to_dict([{"foo": "bar"}])
        self.assertEqual({"0": {"foo": "bar"}}, out)
        self.assertEqual({"f": "b"},
                         stripe.resource.convert_array_to_dict({"f": "b"}))

    def test_raise_on_incorrect_id_type(self):
        for obj in [None, 1, 3.14, dict(), list(), set(), tuple(), object()]:
            self.assertRaises(stripe.error.InvalidRequestError,
                              MyResource.retrieve, obj)

    def test_retrieve_and_update_with_stripe_version(self):
        self.mock_response({
            'id': 'foo',
            'bobble': 'scrobble',
        })

        res = MyUpdateable.retrieve('foo', stripe_version='2017-08-15')

        self.requestor_class_mock.assert_called_with(
            account=None, api_base=None, key=None,
            api_version='2017-08-15'
        )

        res.bobble = 'new_scrobble'
        res.save()

        self.requestor_class_mock.assert_called_with(
            account=None, api_base=None, key='reskey',
            api_version='2017-08-15'
        )


class SingletonAPIResourceTests(StripeApiTestCase):

    def test_retrieve(self):
        self.mock_response({
            'single': 'ton'
        })
        res = MySingleton.retrieve()

        self.requestor_mock.request.assert_called_with(
            'get', '/v1/mysingleton', {}, None)

        self.assertEqual('ton', res.single)


class NestedResourceClassMethodsTests(StripeApiTestCase):
    @stripe.resource.nested_resource_class_methods(
        'nested',
        operations=['create', 'retrieve', 'update', 'delete', 'list']
    )
    class MainResource(stripe.resource.APIResource):
        pass

    def test_create_nested(self):
        self.mock_response({
            'id': 'nested_id',
            'object': 'nested',
            'foo': 'bar',
        })
        nested_resource = self.MainResource.create_nested('id', foo='bar')
        self.requestor_mock.request.assert_called_with(
            'post', '/v1/mainresources/id/nesteds', {'foo': 'bar'}, None)
        self.assertEqual('bar', nested_resource.foo)

    def test_retrieve_nested(self):
        self.mock_response({
            'id': 'nested_id',
            'object': 'nested',
            'foo': 'bar',
        })
        nested_resource = self.MainResource.retrieve_nested('id', 'nested_id')
        self.requestor_mock.request.assert_called_with(
            'get', '/v1/mainresources/id/nesteds/nested_id', {}, None)
        self.assertEqual('bar', nested_resource.foo)

    def test_modify_nested(self):
        self.mock_response({
            'id': 'nested_id',
            'object': 'nested',
            'foo': 'baz',
        })
        nested_resource = self.MainResource.modify_nested('id', 'nested_id',
                                                          foo='baz')
        self.requestor_mock.request.assert_called_with(
            'post', '/v1/mainresources/id/nesteds/nested_id', {'foo': 'baz'},
            None)
        self.assertEqual('baz', nested_resource.foo)

    def test_delete_nested(self):
        self.mock_response({
            'id': 'nested_id',
            'object': 'nested',
            'deleted': True,
        })
        nested_resource = self.MainResource.delete_nested('id', 'nested_id')
        self.requestor_mock.request.assert_called_with(
            'delete', '/v1/mainresources/id/nesteds/nested_id', {}, None)
        self.assertEqual(True, nested_resource.deleted)

    def test_list_nesteds(self):
        self.mock_response({
            'object': 'list',
            'data': [],
        })
        nested_resource = self.MainResource.list_nesteds('id')
        self.requestor_mock.request.assert_called_with(
            'get', '/v1/mainresources/id/nesteds', {}, None)
        self.assertTrue(isinstance(nested_resource.data, list))



import warnings

import stripe
from stripe.test.helper import StripeResourceTest


class SourceTest(StripeResourceTest):

    def test_retrieve_resource(self):
        stripe.Source.retrieve("src_foo")
        self.requestor_mock.request.assert_called_with(
            'get',
            '/v1/sources/src_foo',
            {},
            None
        )

    def test_create_source(self):
        stripe.Source.create(type="bitcoin", amount=1000, currency="usd",
                             owner={"email": "jenny.rosen@example.com"})
        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/sources',
            {
                'type': 'bitcoin',
                'amount': 1000,
                'currency': 'usd',
                'owner': {'email': 'jenny.rosen@example.com'}
            },
            None
        )

    def test_update_source(self):
        source = stripe.Source.construct_from({
            'id': 'src_foo',
            'type': 'card',
            'metadata': {},
        }, 'api_key')
        source.metadata['foo'] = 'bar'
        source.save()

        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/sources/src_foo',
            {
                'metadata': {'foo': 'bar'},
            },
            None
        )

    def test_detach_source_unattached(self):
        source = stripe.Source.construct_from({
            'id': 'src_foo',
        }, 'api_key')
        self.assertRaises(NotImplementedError, source.detach)

    def test_detach_source_attached(self):
        source = stripe.Source.construct_from({
            'id': 'src_foo',
            'customer': 'cus_bar',
        }, 'api_key')
        source.detach()

        self.requestor_mock.request.assert_called_with(
            'delete',
            '/v1/customers/cus_bar/sources/src_foo',
            {},
            None
        )

    def test_delete_source(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')

            source = stripe.Source.construct_from({
                'id': 'src_foo',
                'customer': 'cus_bar',
            }, 'api_key')
            source.delete()

            self.assertEqual(1, len(w))
            self.assertEqual(w[0].category, DeprecationWarning)

    def test_verify_source(self):
        source = stripe.Source.construct_from({
            'id': 'src_foo',
            'type': 'ach_debit'
        }, 'api_key')
        source.verify(values=[32, 45])
        self.requestor_mock.request.assert_called_with(
            'post',
            '/v1/sources/src_foo/verify',
            {
                'values': [32, 45],
            },
            None
        )
